#include <Arduino_APDS9960.h>
#include <Arduino_LSM9DS1.h>
#include <ArduinoBLE.h>

BLEService bleMidiService("03B80E5A-EDE8-4B33-A751-6CE34EC4C700");
BLECharacteristic bleMidiCharacteristic("7772E5DB-3868-4112-A1A9-F2669D106BF3", BLERead | BLEWrite | BLEWriteWithoutResponse | BLENotify, 5);

// カラーセンサのホワイトバランス補正値
float normalize_val[3] = { 1.0f, 1.0f, 1.0f };

void setup() {
  Serial.begin(9600);

  // PCが接続されるまで待つ処理。
  // モバイルバッテリー等で動作させる場合には以下をコメントアウトする。
  //while (!Serial);

  Serial.println("CQ-BLE MIDI Device started");

  APDS.begin();
  IMU.begin();
  BLE.begin();

  // BLEでアドバタイズするデバイスの名称を設定する
  BLE.setLocalName("CQ-BLE MIDI Device");
  BLE.setDeviceName("CQ-BLE MIDI Device");
  // BLEサービスのUUIDをセットする
  BLE.setAdvertisedService(bleMidiService);

  bleMidiService.addCharacteristic(bleMidiCharacteristic);
  BLE.addService(bleMidiService);

  bleMidiCharacteristic.setValue(0);

  // BLEアドバタイズを開始する
  BLE.advertise();

  // 光源(RGB LED)が接続されている端子を出力設定にする
  pinMode(22, OUTPUT);
  pinMode(23, OUTPUT);
  pinMode(24, OUTPUT);

  // ホワイトバランス処理を呼び出す
  checkWhiteBalance();

  delay(1000);

  // 光源をONにする
  setColorLed(true);
}

/*
   光源(RGB LED)の点灯・消灯制御
*/
void setColorLed(bool enable)
{
  if (enable) {
    digitalWrite(22, LOW);
    digitalWrite(23, LOW);
    digitalWrite(24, LOW);
  } else {
    digitalWrite(22, HIGH);
    digitalWrite(23, HIGH);
    digitalWrite(24, HIGH);
  }
}

/*
   ホワイトバランス補正値の計算
*/
void checkWhiteBalance(void)
{
  const int averageTimes = 10;
  int32_t sum_r, sum_g, sum_b;
  int32_t avg_r, avg_g, avg_b;
  int32_t r, g, b;
  int cnt;

  setColorLed(true);

  delay(1000);

  sum_r = sum_g = sum_b = 0;

  // カラーセンサーからの値取得と平均値算出
  for (cnt = 0; cnt < averageTimes;) {
    if (APDS.colorAvailable()) {
      APDS.readColor((int&)r, (int&)g, (int&)b);
      sum_r += r;
      sum_g += g;
      sum_b += b;
      cnt++;
      Serial.printf("%d, %d, %d\n", r, g, b);
      delay(200);
    }
  }
  avg_r = sum_r / averageTimes;
  avg_g = sum_g / averageTimes;
  avg_b = sum_b / averageTimes;

  // 補正値計算
  normalize_val[0] = 2048.0f / avg_r;
  normalize_val[1] = 2048.0f / avg_g;
  normalize_val[2] = 2048.0f / avg_b;

  Serial.printf("Average : %d, %d, %d\n", avg_r, avg_g, avg_b);

  setColorLed(false);
}

/*
   カラーセンサーから得られた値にホワイトバランス補正値を適用する
*/
void correctWhiteBalance(int32_t r1, int32_t g1, int32_t b1, int32_t& r2, int32_t& g2, int32_t& b2)
{
  r2 = static_cast<int32_t>(normalize_val[0] * r1);
  g2 = static_cast<int32_t>(normalize_val[1] * g1);
  b2 = static_cast<int32_t>(normalize_val[2] * b1);

  r2 = r2 > 4095 ? 4095 : r2;
  g2 = g2 > 4095 ? 4095 : g2;
  b2 = b2 > 4095 ? 4095 : b2;
}

/*
   RGB値からHSV値を算出し、色コードを取得する
*/
uint8_t detectColor(int32_t r, int32_t g, int32_t b)
{
  int32_t val_max;
  int32_t val_min;
  int32_t h;
  int32_t s;
  int32_t v;

  val_max = max(max(r, g), b);
  val_min = min(min(r, g), b);

  // Hue
  // 範囲 : 0 - 360

  if ((r == g) && (r == b)) {
    h = 0;
  } else {
    if ((r > g) && (r > b)) {
      h = 60 * (g - b) / (val_max - val_min);
    } else if ((g > r) && (g > b)) {
      h = 60 * (b - r) / (val_max - val_min) + 120;
    } else {
      h = 60 * (r - g) / (val_max - val_min) + 240;
    }
  }
  if (h < 0) {
    h += 360;
  }

  // Saturation
  // 範囲 : 0 - 100

  if (val_max != 0) {
    s = ((val_max - val_min) * 100) / val_max;
  } else {
    s = 0;
  }

  // Value of Brightness
  // 範囲 : 0 - 100
  v = val_max * 100 / 2048;

  // ここまででHSV値が取得できたので、色コード変換関数
  // を呼び出す
  uint8_t colorCode = getColorFromHSV(h, s, v);

  return colorCode;
}

/*
   HSV値から色コードを得る
    色コード
     0 : 黒, 1 : 赤, 2 : 黃, 3 : 緑, 4 : 水,5 : 青, 6 : 紫, 7 : 白
*/
uint8_t getColorFromHSV(int32_t h, int32_t s, int32_t v)
{
  const int32_t threshS = 22; // 無彩色または有彩色の判定閾値
  const int32_t threshV = 10; // 白か黒の判定閾値

  if (s < threshS) {
    // 無彩色
    if (v < threshV) {
      return 0;
    } else {
      return 7;
    }
  } else {
    // 有彩色
    uint8_t c;
    c = (h + 90) / 60;
    c = (c == 7) ? 1 : c;
    return c;
  }
}

void loop() {
  static uint32_t detectedColor;

  BLE.poll();   // BLEの周期処理

  /*
     カラーセンサーを使用してカラーコードを取得する処理
  */
  {
    int32_t r, g, b;
    int32_t r2, g2, b2;

    if (APDS.colorAvailable()) {
      APDS.readColor((int&)r, (int&)g, (int&)b);
      correctWhiteBalance(r, g, b, r2, g2, b2);
      detectedColor = detectColor(r2, g2, b2);
    }
  }

  /*
     ジャイロセンサーを使用してドラムスティックを叩いたタイミングを取得し、
     BLE MIDIメッセージを送信する処理
  */
  {
    static float prev_gy;
    static uint32_t prevHitMillis;
    float gx, gy, gz;
    uint32_t currMillis;
    const float threshDGY = 40.0; // 叩く強さの判定閾値
    const float gainDGY = 0.4;    // 叩く強さのゲイン値

    if (IMU.gyroscopeAvailable()) {
      IMU.readGyroscope(gx, gy, gz);
      //Serial.println(gy - prev_gy);

      float dgy = gy - prev_gy;
      currMillis = millis();

      if (((prev_gy > threshDGY) && (dgy < -threshDGY)) &&
          (prevHitMillis + 50 < currMillis)) {

        // MIDIメッセージに載せるVelocity値(1~127)にΔgyまたはprev_gyをマッピング
        uint32_t velocity = static_cast<int32_t>((-dgy    - threshDGY) * gainDGY); // dgyを利用する場合。
        //      uint32_t velocity = static_cast<int32_t>((prev_gy -threshDGY) *gainDGY);  // 前角速度を利用する場合。叩く強さに忠実

        velocity = velocity > 127 ? 127 : velocity;
        prevHitMillis = currMillis;

        Serial.printf("Hit %+5.0f, %+5.0f, %d, %3d\n", prev_gy, dgy, detectedColor, velocity);
        sendNote(detectedColor, static_cast<uint8_t>(velocity));
      }
    }

    prev_gy = gy;
  }
}

/*
   BLE MIDIメッセージの送信
*/
void sendNote(uint32_t note, uint8_t velocity)
{
  uint32_t currentMillis;
  static byte noteOn[] = { 0x80, 0x80, 0x91, 0x00, 0x127 };
  static byte noteOff[] = { 0x80, 0x80, 0x81, 0x00, 0x127 };
  const uint8_t notes[] = {
    36,   // Bass drum
    38,   // Snare drum
    42,   // Closed Hi-Hat
    46,   // Open Hi-Hat
    45,   // Low Tom
    48,   // Hi-Mid Tom
    50,   // Hi Tom
    49,   // Crash cymbal
  };

  Serial.println(note);

  noteOn[3] = notes[note];
  noteOff[3] = notes[note];

  noteOn[4] = velocity;

  currentMillis = millis();
  noteOn[0] = ((currentMillis >> 7) & 0x3f) | 0x80;
  noteOn[1] = (currentMillis & 0x7f) | 0x80;
  bleMidiCharacteristic.setValue(noteOn, 5);

  delay(1);

  currentMillis = millis();
  noteOff[0] = ((currentMillis >> 7) & 0x3f) | 0x80;
  noteOff[1] = (currentMillis & 0x7f) | 0x80;
  bleMidiCharacteristic.setValue(noteOff, 5);
}
